import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary import constants
from aviary.subsystems.mass.gasp_based.fixed import (ControlMass,
                                                     ElectricAugmentationMass,
                                                     EngineMass,
                                                     FixedMassGroup, GearMass,
                                                     HighLiftMass,
                                                     MassParameters,
                                                     PayloadMass, TailMass)
from aviary.utils.aviary_values import AviaryValues, get_keys
from aviary.variable_info.options import get_option_defaults
from aviary.variable_info.variables import Aircraft, Mission


class MassParametersTestCase1(
    unittest.TestCase
):  # this is large single aisle 1 v3 bug fixed test case
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=True, units='unitless')
        options.set_val(Aircraft.Engine.NUM_FUSELAGE_ENGINES, val=0)

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "parameters",
            MassParameters(
                aviary_options=options
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SWEEP, val=25, units="deg"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=118.8, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            "max_mach", val=0.9, units="unitless"
        )  # bug fixed value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MATERIAL_FACTOR], 1.2203729275531838, tol
        )  # bug fixed value
        assert_near_equal(self.prob["c_strut_braced"], 1, tol)  # bug fixed value
        assert_near_equal(self.prob["c_gear_loc"], 1, tol)  # bug fixed value
        # bug fixed value
        assert_near_equal(self.prob[Aircraft.Engine.POSITION_FACTOR], 0.95, tol)
        assert_near_equal(
            self.prob["half_sweep"], 0.3947081519145335, tol
        )  # bug fixed value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class MassParametersTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Engine.NUM_FUSELAGE_ENGINES, val=2, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "parameters",
            MassParameters(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=117.8, units="ft"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            "max_mach", val=0.72, units="unitless"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MATERIAL_FACTOR], 1.2213063198183813, tol
        )  # not actual bug fixed value
        assert_near_equal(self.prob["c_strut_braced"], 1, tol)
        assert_near_equal(
            self.prob["c_gear_loc"], 0.95, tol
        )  # not actual bug fixed value
        # not actual bug fixed value
        assert_near_equal(self.prob[Aircraft.Engine.POSITION_FACTOR], 1, tol)
        assert_near_equal(self.prob["half_sweep"], 0.3947081519145335, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class MassParametersTestCase3(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES, val=3, units='unitless')
        options.set_val(Aircraft.Engine.NUM_FUSELAGE_ENGINES, val=0, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "parameters",
            MassParameters(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=117.8, units="ft"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            "max_mach", val=0.72, units="unitless"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MATERIAL_FACTOR], 1.2213063198183813, tol
        )  # not actual bug fixed value
        assert_near_equal(self.prob["c_strut_braced"], 1, tol)
        assert_near_equal(
            self.prob["c_gear_loc"], 0.95, tol
        )  # not actual bug fixed value
        assert_near_equal(
            self.prob[Aircraft.Engine.POSITION_FACTOR], 0.98, tol
        )  # not actual bug fixed value
        assert_near_equal(self.prob["half_sweep"], 0.3947081519145335, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class MassParametersTestCase4(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES, val=4, units='unitless')
        options.set_val(Aircraft.Engine.NUM_FUSELAGE_ENGINES, val=0, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "parameters",
            MassParameters(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=117.8, units="ft"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            "max_mach", val=0.72, units="unitless"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MATERIAL_FACTOR], 1.2213063198183813, tol
        )  # not actual bug fixed value
        assert_near_equal(self.prob["c_strut_braced"], 1, tol)
        assert_near_equal(
            self.prob["c_gear_loc"], 0.95, tol
        )  # not actual bug fixed value
        assert_near_equal(
            self.prob[Aircraft.Engine.POSITION_FACTOR], 0.95, tol
        )  # not actual bug fixed value
        assert_near_equal(self.prob["half_sweep"], 0.3947081519145335, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class MassParametersTestCase5(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES, val=4, units='unitless')
        options.set_val(Aircraft.Engine.NUM_FUSELAGE_ENGINES, val=0, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "parameters",
            MassParameters(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(Aircraft.Wing.SWEEP, val=25, units="deg")
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=117.8, units="ft"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            "max_mach", val=0.9, units="unitless"
        )  # not actual bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.Wing.MATERIAL_FACTOR], 1.2213063198183813, tol
        )  # not actual bug fixed value
        assert_near_equal(self.prob["c_strut_braced"], 1, tol)
        assert_near_equal(
            self.prob["c_gear_loc"], 0.95, tol
        )  # not actual bug fixed value
        assert_near_equal(
            self.prob[Aircraft.Engine.POSITION_FACTOR], 0.9, tol
        )  # not actual bug fixed value
        assert_near_equal(self.prob["half_sweep"], 0.3947081519145335, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


# this is the large single aisle 1 V3 test case
class PayloadMassTestCase(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.CrewPayload.PASSENGER_MASS_WITH_BAGS,
                        val=200, units="lbm")  # bug fixed value and original value

        self.prob = om.Problem()
        self.prob.model.add_subsystem("payload", PayloadMass(
            aviary_options=options), promotes=["*"])
        self.prob.model.set_input_defaults(
            Aircraft.CrewPayload.CARGO_MASS, val=10040, units="lbm"
        )  # bug fixed value and original value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 1e-4
        assert_near_equal(
            self.prob[Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS], 36000, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["payload_mass_des"], 36000, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["payload_mass_max"], 46040, tol
        )  # bug fixed value and original value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class ElectricAugmentationTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "aug", ElectricAugmentationMass(aviary_options=get_option_defaults()), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            "motor_power", val=830, units="kW"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "motor_voltage", val=850, units="V"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "max_amp_per_wire", val=260, units="A"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "safety_factor", val=1.0, units="unitless"
        )  # not included in eTTVW v3.6
        self.prob.model.set_input_defaults(
            Aircraft.Electrical.HYBRID_CABLE_LENGTH, val=65.6, units="ft"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "wire_area", val=0.0015, units="ft**2"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "rho_wire", val=565, units="lbm/ft**3"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "battery_energy", val=6077, units="MJ"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "motor_eff", val=0.98, units="unitless"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "inverter_eff", val=0.99, units="unitless"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "transmission_eff", val=0.975, units="unitless"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "battery_eff", val=0.975, units="unitless"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "rho_battery", val=0.5, units="kW*h/kg"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "motor_spec_mass", val=4, units="hp/lbm"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "inverter_spec_mass", val=12, units="kW/kg"
        )  # electrified diff configuration value v3.6
        self.prob.model.set_input_defaults(
            "TMS_spec_mass", val=0.125, units="lbm/kW"
        )  # electrified diff configuration value v3.6
        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob["aug_mass"], 9394.3, 0.0017
        )  # electrified diff configuration value v3.6. Higher tol because num_wires is discrete in GASP and is not in Aviary

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=4e-12, rtol=1e-12)


class EngineTestCase1(unittest.TestCase):  # this is the large single aisle 1 V3 test case
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM,
                        val=False, units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "engine",
            EngineMass(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SPECIFIC, val=0.21366, units="lbm/lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.MASS_SPECIFIC, val=3, units="lbm/ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.SURFACE_AREA, val=339.58, units="ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.PYLON_FACTOR, val=1.25, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.ADDITIONAL_MASS_FRACTION, val=0.14, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Propulsion.MISC_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.WING_LOCATIONS, val=0.35, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "main_gear_mass", val=6384.35, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0.15, units="unitless"
        )  # bug fixed value and original value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_MASS], 12606.0, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS], 3785.0, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob[Aircraft.Engine.ADDITIONAL_MASS], 1765.0/2, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["eng_comb_mass"], 14370.8, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["wing_mounted_mass"], 24446.343040697346, tol
        )  # bug fixed value and original value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-11, rtol=1e-12)


class EngineTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Engine.HAS_PROPELLERS, val=[True], units='unitless')

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "engine",
            EngineMass(aviary_options=options),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SPECIFIC, val=0.21366, units="lbm/lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500.0, units="lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.MASS_SPECIFIC, val=3, units="lbm/ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.SURFACE_AREA, val=339.58, units="ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.PYLON_FACTOR, val=1.25, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.ADDITIONAL_MASS_FRACTION, val=0.14, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Propulsion.MISC_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "prop_mass", val=0, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "aug_mass", val=0, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.WING_LOCATIONS, val=0.35, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "main_gear_mass", val=6384.35, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0.15, units="unitless"
        )  # bug fixed value and original value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_MASS], 12606.0, tol
        )  # note: these are only the right values because this was given a prop mass of zero. This is not a large single aisle test case
        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS], 3785.0, tol
        )  # note: these are only the right values because this was given a prop mass of zero. This is not a large single aisle test case
        assert_near_equal(
            self.prob[Aircraft.Engine.ADDITIONAL_MASS], 1765.0/2, tol
        )  # note: these are only the right values because this was given a prop mass of zero. This is not a large single aisle test case
        assert_near_equal(
            self.prob["eng_comb_mass"], 14370.8, tol
        )  # note: these are only the right values because this was given a prop mass of zero. This is not a large single aisle test case
        assert_near_equal(
            self.prob["prop_mass_all"], 0, tol
        )  # note: these are only the right values because this was given a prop mass of zero. This is not a large single aisle test case
        assert_near_equal(
            self.prob["wing_mounted_mass"], 24446.343040697346, tol
        )  # note: these are only the right values because this was given a prop mass of zero. This is not a large single aisle test case

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=2e-11, rtol=1e-12)


class TailTestCase(unittest.TestCase):  # this is the large single aisle 1 V3 test case
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "tail", TailMass(aviary_options=get_option_defaults()), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.SWEEP, val=0, units="rad"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.SPAN, val=28.22, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS_COEFFICIENT, val=0.232, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.SPAN, val=42.59, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TAIL_HOOK_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS_COEFFICIENT, val=0.289, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=118.8, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=381.8, units="ft**2"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            "min_dive_vel", val=420, units="kn"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_ARM, val=55.1, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ROOT_CHORD, val=13.261162230765065, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=476.8, units="ft**2"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_ARM, val=50.3, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ROOT_CHORD, val=18.762708015981357, units="ft"
        )  # bug fixed value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob["loc_MAC_vtail"], 0.44959578484694906, tol
        )  # bug fixed value
        # bug fixed value
        assert_near_equal(self.prob[Aircraft.HorizontalTail.MASS], 2285, tol)
        assert_near_equal(self.prob[Aircraft.VerticalTail.MASS],
                          2312, tol)  # bug fixed value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-11, rtol=1e-12)


# this is a different configuration with turbofan_23k_1 test case
class HighLiftTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()

        aviary_options = get_option_defaults()
        aviary_options.set_val(Aircraft.Wing.NUM_FLAP_SEGMENTS, val=2)

        self.prob.model.add_subsystem(
            "HL", HighLiftMass(aviary_options=aviary_options), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT, val=1.9, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1764.6, units="ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SLAT_CHORD_RATIO, val=0.1, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FLAP_CHORD_RATIO, val=0.25, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.346, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SLAT_SPAN_RATIO, val=0.9, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.FLAP_SPAN_RATIO, val=0.88, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.LOADING, val=93.1, units="lbf/ft**2"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.11, units="unitless"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=185.8, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.AVG_DIAMETER, val=13.1, units="ft"
        )
        self.prob.model.set_input_defaults(
            Aircraft.Wing.CENTER_CHORD, val=13.979, units="ft"
        )
        self.prob.model.set_input_defaults(
            Mission.Landing.LIFT_COEFFICIENT_MAX, val=2.3648, units="unitless"
        )

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        # bug fixed value
        assert_near_equal(self.prob[Aircraft.Wing.HIGH_LIFT_MASS], 4829.6, tol)

        partial_data = self.prob.check_partials(
            out_stream=None, method="cs", show_only_incorrect=True)
        assert_check_partials(partial_data, atol=5e-10, rtol=1e-12)


# this is the large single aisle 1 V3 test case
class ControlMassTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem("payload", ControlMass(
            aviary_options=get_option_defaults()), promotes=["*"])

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SURFACE_CONTROL_MASS_COEFFICIENT, val=0.95, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1392.1, units="ft**2"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.951, units="unitless"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            "min_dive_vel", val=420, units="kn"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Design.COCKPIT_CONTROL_MASS_COEFFICIENT, val=16.5, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS, val=0, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.COCKPIT_CONTROL_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.CONTROL_MASS_INCREMENT, val=0, units="lbm"
        )  # bug fixed value and original value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        # bug fixed value
        assert_near_equal(self.prob[Aircraft.Controls.TOTAL_MASS], 3945, tol)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-11, rtol=1e-12)


class GearTestCase1(unittest.TestCase):  # this is the large single aisle 1 V3 test case
    def setUp(self):

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "payload", GearMass(aviary_options=get_option_defaults()), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MASS_COEFFICIENT, val=0.04, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_MASS_COEFFICIENT, val=0.85, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.CLEARANCE_RATIO, val=0.2, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.AVG_DIAMETER, val=7.35, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MOUNTING_TYPE, val=0, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.LandingGear.TOTAL_MASS], 7511, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["main_gear_mass"], 6384.35, tol
        )  # bug fixed value and original value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=3e-11, rtol=1e-12)


class GearTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "payload", GearMass(aviary_options=options), promotes=["*"]
        )

        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MASS_COEFFICIENT, val=0.04, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_MASS_COEFFICIENT, val=0.85, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MOUNTING_TYPE, val=0.1, units="unitless")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob[Aircraft.LandingGear.TOTAL_MASS], 7016, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["main_gear_mass"], 5963.6, tol
        )  # not actual GASP value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


# this is the large single aisle 1 V3 test case
class FixedMassGroupTestCase1(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.Electrical.HAS_HYBRID_SYSTEM,
                        val=False, units='unitless')
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.CrewPayload.PASSENGER_MASS_WITH_BAGS,
                        val=200, units="lbm")  # bug fixed value and original value

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            FixedMassGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=118.8, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "min_dive_vel", val=420, units="kn"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1392.1, units="ft**2"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SWEEP, val=25, units="deg"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "max_mach", val=0.9, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.CrewPayload.CARGO_MASS, val=10040, units="lbm"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.SWEEP, val=0, units="rad"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.SPAN, val=28.22, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS_COEFFICIENT, val=0.232, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.SPAN, val=42.59, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TAIL_HOOK_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS_COEFFICIENT, val=0.289, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=381.8, units="ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_ARM, val=55.1, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ROOT_CHORD, val=13.261162230765065, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=476.8, units="ft**2"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_ARM, val=50.3, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ROOT_CHORD, val=18.762708015981357, units="ft"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT, val=1.9, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Mission.Landing.LIFT_COEFFICIENT_MAX, val=2.966, units="unitless"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SURFACE_CONTROL_MASS_COEFFICIENT, val=0.95, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.951, units="unitless"
        )  # bug fixed value
        self.prob.model.set_input_defaults(
            Aircraft.Design.COCKPIT_CONTROL_MASS_COEFFICIENT, val=16.5, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS, val=0, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.COCKPIT_CONTROL_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.CONTROL_MASS_INCREMENT, val=0, units="lbm"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MASS_COEFFICIENT, val=0.04, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_MASS_COEFFICIENT, val=0.85, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.CLEARANCE_RATIO, val=0.2, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.AVG_DIAMETER, val=7.35, units="ft"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SPECIFIC, val=0.21366, units="lbm/lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500, units="lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.MASS_SPECIFIC, val=3, units="lbm/ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.SURFACE_AREA, val=339.58, units="ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.PYLON_FACTOR, val=1.25, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.ADDITIONAL_MASS_FRACTION, val=0.14, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.WING_LOCATIONS, val=0.35, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0.15, units="unitless"
        )  # bug fixed value and original value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob["main_gear_mass"], 6384.35, tol
        )  # bug fixed value and original value

        assert_near_equal(
            self.prob[Aircraft.Wing.MATERIAL_FACTOR], 1.2203729275531838, tol
        )  # bug fixed value
        assert_near_equal(
            self.prob["c_strut_braced"], 1, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["c_gear_loc"], 1, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob[Aircraft.Engine.POSITION_FACTOR], 0.95, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["half_sweep"], 0.3947081519145335, tol
        )  # bug fixed value and original value

        assert_near_equal(
            self.prob[Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS], 36000, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["payload_mass_des"], 36000, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["payload_mass_max"], 46040, tol
        )  # bug fixed value and original value

        assert_near_equal(
            self.prob["tail.loc_MAC_vtail"], 0.44959578484694906, tol
        )  # bug fixed value
        # bug fixed value
        assert_near_equal(self.prob[Aircraft.HorizontalTail.MASS], 2285, tol)
        assert_near_equal(self.prob[Aircraft.VerticalTail.MASS],
                          2312, tol)  # bug fixed value
        # bug fixed value
        assert_near_equal(self.prob[Aircraft.Wing.HIGH_LIFT_MASS], 4082.1, tol)

        # bug fixed value
        assert_near_equal(self.prob[Aircraft.Controls.TOTAL_MASS], 3945, tol)

        assert_near_equal(
            self.prob[Aircraft.LandingGear.TOTAL_MASS], 7511, tol
        )  # bug fixed value and original value

        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_MASS], 12606, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS], 3785, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob[Aircraft.Engine.ADDITIONAL_MASS], 1765/2, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["eng_comb_mass"], 14370.8, tol
        )  # bug fixed value and original value
        assert_near_equal(
            self.prob["wing_mounted_mass"], 24446.343040697346, tol
        )  # bug fixed value and original value

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=3e-11, rtol=1e-12)


class FixedMassGroupTestCase2(unittest.TestCase):
    def setUp(self):

        options = get_option_defaults()
        options.set_val(Aircraft.CrewPayload.NUM_PASSENGERS, val=180, units='unitless')
        options.set_val(Aircraft.Engine.NUM_FUSELAGE_ENGINES, val=2, units='unitless')
        options.set_val(Aircraft.Engine.HAS_PROPELLERS, val=[True], units='unitless')
        options.set_val(Aircraft.Wing.HAS_STRUT, val=True, units='unitless')
        options.set_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED,
                        val=False, units='unitless')
        options.set_val(Aircraft.CrewPayload.PASSENGER_MASS_WITH_BAGS,
                        val=200, units="lbm")  # bug fixed value and original value

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "group",
            FixedMassGroup(
                aviary_options=options,
            ),
            promotes=["*"],
        )

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SPAN, val=117.8, units="ft"
        )  # original GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.MOUNTING_TYPE, val=.1, units="unitless")
        self.prob.model.set_input_defaults(
            Mission.Design.GROSS_MASS, val=175400, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "min_dive_vel", val=420, units="kn"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.AREA, val=1370.3, units="ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SWEEP, val=25, units="deg"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "max_mach", val=0.72, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, val=10/117.8, units='unitless'
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.CrewPayload.CARGO_MASS, val=10040, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.TAPER_RATIO, val=0.801, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ASPECT_RATIO, val=1.67, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.SWEEP, val=0.1, units="rad"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.SPAN, val=28, units="ft"
        )  # original GASP value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MASS_COEFFICIENT, val=0.232, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Fuselage.LENGTH, val=129.4, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.SPAN, val=42.25, units="ft"
        )  # original GASP value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.TAIL_HOOK_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.TAPER_RATIO, val=0.352, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MASS_COEFFICIENT, val=0.289, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.AREA, val=375.9, units="ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.MOMENT_ARM, val=54.7, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.ROOT_CHORD, val=13.16130387591471, units="ft"
        )  # original GASP value
        self.prob.model.set_input_defaults(
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION, val=0, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.AREA, val=469.3, units="ft**2"
        )  # original GASP value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.MOMENT_ARM, val=49.9, units="ft"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.THICKNESS_TO_CHORD, val=0.12, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.VerticalTail.ROOT_CHORD, val=18.61267549773935, units="ft"
        )  # original GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT, val=1.9, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Mission.Landing.LIFT_COEFFICIENT_MAX, val=2.817, units="unitless"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            Aircraft.Wing.SURFACE_CONTROL_MASS_COEFFICIENT, val=0.95, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR, val=3.893, units="unitless")
        self.prob.model.set_input_defaults(
            Aircraft.Design.COCKPIT_CONTROL_MASS_COEFFICIENT, val=16.5, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS, val=0, units="lbm"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.COCKPIT_CONTROL_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Controls.CONTROL_MASS_INCREMENT, val=0, units="lbm"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MASS_COEFFICIENT, val=0.04, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_MASS_COEFFICIENT, val=0.85, units="unitless"
        )  # bug fixed value and original value

        self.prob.model.set_input_defaults(
            "augmentation.motor_power", val=200, units="kW"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.motor_voltage", val=50, units="V"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.max_amp_per_wire", val=50, units="A"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.safety_factor", val=1.33, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            Aircraft.Electrical.HYBRID_CABLE_LENGTH, val=200, units="ft"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.wire_area", val=0.0015, units="ft**2"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.rho_wire", val=1, units="lbm/ft**3"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.battery_energy", val=1, units="MJ"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.motor_eff", val=1, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.inverter_eff", val=1, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.transmission_eff", val=1, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.battery_eff", val=1, units="unitless"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.rho_battery", val=200, units="kW*h/kg"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.motor_spec_mass", val=10, units="hp/lbm"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.inverter_spec_mass", val=10, units="kW/kg"
        )  # not actual GASP value
        self.prob.model.set_input_defaults(
            "augmentation.TMS_spec_mass", val=0.125, units="lbm/kW"
        )  # electrified diff configuration value v3.6

        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SPECIFIC, val=0.21366, units="lbm/lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.SCALED_SLS_THRUST, val=29500, units="lbf"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.MASS_SPECIFIC, val=3, units="lbm/ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Nacelle.SURFACE_AREA, val=339.58, units="ft**2"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.PYLON_FACTOR, val=1.25, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.ADDITIONAL_MASS_FRACTION, val=0.14, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.MASS_SCALER, val=1, units="unitless"
        )  # bug fixed value and original value
        # self.prob.model.set_input_defaults(
        #     Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1, units="unitless"
        # )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.Engine.WING_LOCATIONS, val=0.35, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            Aircraft.LandingGear.MAIN_GEAR_LOCATION, val=0.15, units="unitless"
        )  # bug fixed value and original value
        self.prob.model.set_input_defaults(
            "engine.prop_mass", val=0, units="lbm"
        )  # bug fixed value and original value

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_case1(self):

        self.prob.run_model()

        tol = 5e-4
        assert_near_equal(
            self.prob["main_gear_mass"], 5963.6, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["aug_mass"], 228.51036478, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Wing.MATERIAL_FACTOR], 1.2213063198183813, tol)
        assert_near_equal(
            self.prob["c_strut_braced"], 0.9928, tol
        )  # not actual GASP value
        assert_near_equal(self.prob["c_gear_loc"], 1, tol)  # not actual GASP value
        # not actual GASP value
        assert_near_equal(self.prob[Aircraft.Engine.POSITION_FACTOR], 1, tol)
        assert_near_equal(
            self.prob["half_sweep"], 0.3947081519145335, tol
        )  # bug fixed and original value

        assert_near_equal(
            self.prob[Aircraft.CrewPayload.PASSENGER_PAYLOAD_MASS], 36000, tol
        )  # bug fixed and original value
        assert_near_equal(
            self.prob["payload_mass_des"], 36000, tol
        )  # bug fixed and original value
        assert_near_equal(
            self.prob["payload_mass_max"], 46040, tol
        )  # bug fixed and original value

        assert_near_equal(
            self.prob["tail.loc_MAC_vtail"], 1.799, tol
        )  # not actual GASP value
        # original GASP value
        assert_near_equal(self.prob[Aircraft.HorizontalTail.MASS], 2275, tol)
        assert_near_equal(self.prob[Aircraft.VerticalTail.MASS],
                          2297, tol)  # original GASP value
        # original GASP value
        assert_near_equal(self.prob[Aircraft.Wing.HIGH_LIFT_MASS], 4162.1, tol)

        # original GASP value
        assert_near_equal(self.prob[Aircraft.Controls.TOTAL_MASS], 3895, tol)

        assert_near_equal(
            self.prob[Aircraft.LandingGear.TOTAL_MASS], 7016, tol
        )  # not actual GASP value

        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_MASS], 12606, tol
        )  # bug fixed and original value
        assert_near_equal(
            self.prob[Aircraft.Propulsion.TOTAL_ENGINE_POD_MASS], 3785, tol
        )  # bug fixed and original value
        assert_near_equal(
            self.prob[Aircraft.Engine.ADDITIONAL_MASS], 1765/2, tol
        )  # bug fixed and original value
        assert_near_equal(
            self.prob["eng_comb_mass"], 14599.28196478, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["wing_mounted_mass"], 24027.6, tol
        )  # not actual GASP value
        assert_near_equal(
            self.prob["engine.prop_mass_all"], 0, tol
        )  # bug fixed and original value

        partial_data = self.prob.check_partials(
            out_stream=None, method="cs", form="central",
        )
        assert_check_partials(partial_data, atol=3e-10, rtol=1e-12)


class FixedMassGroupTestCase3(unittest.TestCase):
    # Tests partials calculations in FixedMassGroup using as complete code coverage as possible.

    def _run_case(self, data):

        prob = om.Problem()
        prob.model.add_subsystem(
            'fixed_mass',
            FixedMassGroup(aviary_options=data),
            promotes=['*'],
        )
        prob.setup(force_alloc_complex=True)

        for key in get_keys(data):
            val, units = data.get_item(key)
            try:
                prob.set_val(key, val, units)
            except:
                pass

        prob.run_model()

        partial_data = prob.check_partials(out_stream=None, method='cs')
        assert_check_partials(partial_data, atol=1e-7, rtol=1e-12)

    def setUp(self):
        import aviary.subsystems.mass.gasp_based.fixed as fixed

        # Set GRAV_ENGLISH_LBM = 1.1 to find errors that aren't
        # caught when GRAV_ENGLISH_LBM = 1 and is misplaced.
        constants.GRAV_ENGLISH_LBM = 1.1
        fixed.GRAV_ENGLISH_LBM = 1.1

    def tearDown(self):
        import aviary.subsystems.mass.gasp_based.fixed as fixed
        constants.GRAV_ENGLISH_LBM = 1.0
        fixed.GRAV_ENGLISH_LBM = 1.0

    def test_case1(self):

        data = AviaryValues({
            Aircraft.Engine.NUM_ENGINES: (2, 'unitless'),
            Aircraft.Propulsion.TOTAL_NUM_ENGINES: (2, 'unitless'),
            Aircraft.Design.SMOOTH_MASS_DISCONTINUITIES: (False, 'unitless'),
            Aircraft.Engine.NUM_FUSELAGE_ENGINES: (0, 'unitless'),
            Aircraft.CrewPayload.NUM_PASSENGERS: (150, 'unitless'),
            Aircraft.Electrical.HAS_HYBRID_SYSTEM: (False, 'unitless'),
            Aircraft.Engine.HAS_PROPELLERS: ([False], 'unitless'),
            Aircraft.Wing.FLAP_TYPE: ('plain', 'unitless'),
            Aircraft.Wing.SWEEP: (30.0, 'deg'),
            Aircraft.Wing.MOUNTING_TYPE: (0, 'unitless'),
            Aircraft.Wing.TAPER_RATIO: (0.25, 'unitless'),
            Aircraft.Wing.ASPECT_RATIO: (11.0, 'unitless'),
            Aircraft.Wing.SPAN: (100.0, 'ft'),
            'max_mach': (0.9, 'unitless'),
            Aircraft.Strut.ATTACHMENT_LOCATION: (10.0, 'ft'),
            Aircraft.LandingGear.MAIN_GEAR_LOCATION: (0.2, 'unitless'),
            Aircraft.CrewPayload.PASSENGER_MASS_WITH_BAGS: (200.0, 'lbm'),
            Aircraft.CrewPayload.CARGO_MASS: (10040.0, 'lbm'),
            'wire_area': (0.0015, 'ft**2'),
            'rho_wire': (1.1, 'lbm/ft**3'),
            'battery_energy': (0.0, 'MJ'),
            'motor_eff': (1.5, 'unitless'),
            'inverter_eff': (0.95, 'unitless'),
            'transmission_eff': (0.96, 'unitless'),
            'battery_eff': (0.97, 'unitless'),
            'rho_battery': (210.0, 'MJ/lb'),
            'motor_spec_mass': (10.0, 'hp/lbm'),
            'inverter_spec_mass': (10.5, 'kW/lbm'),
            'TMS_spec_mass': (10.6, 'lbm/kW'),
            Aircraft.Engine.MASS_SPECIFIC: (0.21366, 'lbm/lbf'),
            Aircraft.Engine.SCALED_SLS_THRUST: (4000.0, 'lbf'),
            Aircraft.Nacelle.MASS_SPECIFIC: (3.0, 'lbm/ft**2'),
            Aircraft.Nacelle.SURFACE_AREA: (5.0, 'ft**2'),
            Aircraft.Engine.PYLON_FACTOR: (1.25, 'unitless'),
            Aircraft.Engine.ADDITIONAL_MASS_FRACTION: (0.14, 'unitless'),
            Aircraft.Engine.MASS_SCALER: (1.05, 'unitless'),
            Aircraft.Propulsion.MISC_MASS_SCALER: (1.06, 'unitless'),
            Aircraft.Engine.WING_LOCATIONS: (0.35, 'unitless'),
            'prop_mass': (0.5, 'lbm'),
            Aircraft.VerticalTail.TAPER_RATIO: (0.26, 'unitless'),
            Aircraft.VerticalTail.ASPECT_RATIO: (5.0, 'unitless'),
            Aircraft.VerticalTail.SWEEP: (25.0, 'deg'),
            Aircraft.VerticalTail.SPAN: (20.0, 'ft'),
            Mission.Design.GROSS_MASS: (152000.0, 'lbm'),
            Aircraft.HorizontalTail.MASS_COEFFICIENT: (1.07, 'unitless'),
            Aircraft.Fuselage.LENGTH: (120.0, 'ft'),
            Aircraft.HorizontalTail.SPAN: (14.0, 'ft'),
            Aircraft.LandingGear.TAIL_HOOK_MASS_SCALER: (1.08, 'unitless'),
            Aircraft.HorizontalTail.TAPER_RATIO: (0.35, 'unitless'),
            Aircraft.VerticalTail.MASS_COEFFICIENT: (1.09, 'unitless'),
            Aircraft.HorizontalTail.AREA: (500.0, 'ft**2'),
            'min_dive_vel': (200.0, 'kn'),
            Aircraft.HorizontalTail.MOMENT_ARM: (20.0, 'ft'),
            Aircraft.HorizontalTail.THICKNESS_TO_CHORD: (0.15, 'unitless'),
            Aircraft.HorizontalTail.ROOT_CHORD: (8.0, 'ft'),
            Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION: (0.3, 'unitless'),
            Aircraft.VerticalTail.AREA: (250.0, 'ft**2'),
            Aircraft.VerticalTail.MOMENT_ARM: (6.0, 'ft'),
            Aircraft.VerticalTail.THICKNESS_TO_CHORD: (0.12, 'unitless'),
            Aircraft.VerticalTail.ROOT_CHORD: (7.0, 'ft'),
            Aircraft.Wing.HIGH_LIFT_MASS_COEFFICIENT: (1.10, 'unitless'),
            Aircraft.Wing.AREA: (3500.0, 'ft**2'),
            Aircraft.Wing.NUM_FLAP_SEGMENTS: (2, 'unitless'),
            Aircraft.Wing.SLAT_CHORD_RATIO: (0.15, 'unitless'),
            Aircraft.Wing.FLAP_CHORD_RATIO: (0.3, 'unitless'),
            Aircraft.Wing.SLAT_SPAN_RATIO: (0.9, 'unitless'),
            Aircraft.Wing.FLAP_SPAN_RATIO: (0.65, 'unitless'),
            Aircraft.Wing.LOADING: (128.0, 'lbf/ft**2'),
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT: (0.15, 'unitless'),
            Aircraft.Fuselage.AVG_DIAMETER: (11.0, 'ft'),
            Aircraft.Wing.CENTER_CHORD: (17.0, 'ft'),
            Mission.Landing.LIFT_COEFFICIENT_MAX: (1.8, 'unitless'),
            'density': (0.0023769, 'slug/ft**3'),
            Aircraft.Wing.ULTIMATE_LOAD_FACTOR: (7.0, 'unitless'),
            Aircraft.Design.COCKPIT_CONTROL_MASS_COEFFICIENT: (1.11, 'unitless'),
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS: (200.0, 'lbm'),
            Aircraft.Controls.COCKPIT_CONTROL_MASS_SCALER: (1.12, 'unitless'),
            Aircraft.Wing.SURFACE_CONTROL_MASS_SCALER: (1.13, 'unitless'),
            Aircraft.Controls.STABILITY_AUGMENTATION_SYSTEM_MASS_SCALER: (1.14, 'unitless'),
            Aircraft.Controls.CONTROL_MASS_INCREMENT: (25.0, 'lbm'),
            Aircraft.LandingGear.MASS_COEFFICIENT: (1.15, 'unitless'),
            Aircraft.LandingGear.MAIN_GEAR_MASS_COEFFICIENT: (1.16, 'unitless'),
            Aircraft.Nacelle.CLEARANCE_RATIO: (0.2, 'unitless'),
            Aircraft.Nacelle.AVG_DIAMETER: (7.5, 'ft')
        })

        # Try to cover all if-then branches in fixed.py.
        for flap_type in ['split', 'single_slotted', 'fowler']:
            for has_hybrid in [False, True]:
                for has_prop in [False, True]:
                    for gear_loc in [0.0, 0.05, 0.01]:
                        for fuse_mounted in [False, True]:
                            for num_engines in [2, 4]:
                                num_fuse_eng = num_engines if fuse_mounted else 0
                                data.set_val(
                                    Aircraft.Engine.NUM_FUSELAGE_ENGINES, num_fuse_eng)
                                data.set_val(
                                    Aircraft.Engine.NUM_ENGINES, num_engines)
                                data.set_val(
                                    Aircraft.Propulsion.TOTAL_NUM_ENGINES, num_engines)
                                data.set_val(
                                    Aircraft.LandingGear.MAIN_GEAR_LOCATION, gear_loc)
                                data.set_val(
                                    Aircraft.Wing.FLAP_TYPE, flap_type)
                                data.set_val(
                                    Aircraft.Electrical.HAS_HYBRID_SYSTEM, has_hybrid)
                                data.set_val(
                                    Aircraft.Engine.HAS_PROPELLERS, [has_prop])

                                self._run_case(data)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_GASP, GRAV_ENGLISH_LBM
from aviary.variable_info.enums import AnalysisScheme
from aviary.variable_info.variables import Dynamic


class AccelerationRates(om.ExplicitComponent):

    """
    Compute the TAS rate, distance rate, and mass flow rate for a level flight acceleration phase.

    Equation comes from climb subroutine of GASP code.
    """

    def initialize(self):
        self.options.declare("num_nodes", types=int)
        self.options.declare("analysis_scheme", types=AnalysisScheme, default=AnalysisScheme.COLLOCATION,
                             desc="The analysis method that will be used to close the trajectory; for example collocation or time integration")

    def setup(self):
        analysis_scheme = self.options["analysis_scheme"]
        nn = self.options["num_nodes"]
        arange = np.arange(nn)

        self.add_input(
            Dynamic.Mission.MASS,
            val=np.ones(nn) * 1e6,
            units="lbm",
            desc="total mass of the aircraft",
        )
        self.add_input(
            Dynamic.Mission.DRAG,
            val=np.zeros(nn),
            units="lbf",
            desc="drag on aircraft",
        )
        self.add_input(
            Dynamic.Mission.THRUST_TOTAL,
            val=np.zeros(nn),
            units="lbf",
            desc="total thrust",
        )
        self.add_input(
            Dynamic.Mission.VELOCITY,
            val=np.zeros(nn),
            units="ft/s",
            desc="true air speed",
        )

        self.add_output(
            Dynamic.Mission.VELOCITY_RATE,
            val=np.zeros(nn),
            units="ft/s**2",
            desc="rate of change of true air speed",
        )
        self.add_output(
            Dynamic.Mission.DISTANCE_RATE,
            val=np.zeros(nn),
            units="ft/s",
            desc="rate of change of horizontal distance covered",
        )

        self.declare_partials(
            Dynamic.Mission.VELOCITY_RATE, [
                Dynamic.Mission.MASS, Dynamic.Mission.DRAG, Dynamic.Mission.THRUST_TOTAL], rows=arange, cols=arange)
        self.declare_partials(Dynamic.Mission.DISTANCE_RATE, [
                              Dynamic.Mission.VELOCITY], rows=arange, cols=arange, val=1.)

        if analysis_scheme is AnalysisScheme.SHOOTING:
            self.add_input("t_curr", val=np.ones(nn), desc="time", units="s")
            self.add_output(Dynamic.Mission.ALTITUDE_RATE, val=np.ones(nn),
                            desc="altitude rate", units="ft/s")
            self.add_input(
                Dynamic.Mission.DISTANCE, val=np.ones(nn), desc="distance traveled", units="ft")

    def compute(self, inputs, outputs):
        analysis_scheme = self.options["analysis_scheme"]

        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM
        drag = inputs[Dynamic.Mission.DRAG]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]
        TAS = inputs[Dynamic.Mission.VELOCITY]

        outputs[Dynamic.Mission.VELOCITY_RATE] = (
            GRAV_ENGLISH_GASP / weight) * (thrust - drag)
        outputs[Dynamic.Mission.DISTANCE_RATE] = TAS

        if analysis_scheme is AnalysisScheme.SHOOTING:
            outputs[Dynamic.Mission.ALTITUDE_RATE] = 0

    def compute_partials(self, inputs, J):
        weight = inputs[Dynamic.Mission.MASS] * GRAV_ENGLISH_LBM
        drag = inputs[Dynamic.Mission.DRAG]
        thrust = inputs[Dynamic.Mission.THRUST_TOTAL]

        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.MASS] = \
            -(GRAV_ENGLISH_GASP / weight**2) * (thrust - drag) * GRAV_ENGLISH_LBM
        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.DRAG] = - \
            (GRAV_ENGLISH_GASP / weight)
        J[Dynamic.Mission.VELOCITY_RATE, Dynamic.Mission.THRUST_TOTAL] = GRAV_ENGLISH_GASP / weight



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class FinMass(om.ExplicitComponent):
    '''
    Calculates the mass of the fin(s). The methodology is based on the FLOPS weight
    equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Fins.AREA, val=0.0)
        add_aviary_input(self, Aircraft.Fins.TAPER_RATIO, val=0.0)
        add_aviary_input(self, Aircraft.Fins.MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.Fins.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials("*", "*")

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_fins = aviary_options.get_val(Aircraft.Fins.NUM_FINS)
        if num_fins > 0:
            togw = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM
            area = inputs[Aircraft.Fins.AREA]
            taper_ratio = inputs[Aircraft.Fins.TAPER_RATIO]

            fin_weight = 0.32 * togw**0.3 * area**0.85 * (taper_ratio+0.5) * num_fins
            outputs[Aircraft.Fins.MASS] = fin_weight * \
                inputs[Aircraft.Fins.MASS_SCALER] / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_fins = aviary_options.get_val(Aircraft.Fins.NUM_FINS)
        if num_fins > 0:
            area = inputs[Aircraft.Fins.AREA]
            taper_ratio = inputs[Aircraft.Fins.TAPER_RATIO]
            scalar = inputs[Aircraft.Fins.MASS_SCALER]
            gross_weight = inputs[Mission.Design.GROSS_MASS] * GRAV_ENGLISH_LBM

            area_exp = area**0.85
            gross_weight_exp = gross_weight**0.3

            J[Aircraft.Fins.MASS, Aircraft.Fins.AREA] = (
                (0.272 * num_fins * scalar * (taper_ratio + 0.5) * gross_weight_exp)
                / area**0.15) / GRAV_ENGLISH_LBM
            J[Aircraft.Fins.MASS, Aircraft.Fins.TAPER_RATIO] = \
                0.32 * area_exp * num_fins * scalar * gross_weight_exp / GRAV_ENGLISH_LBM
            J[Aircraft.Fins.MASS, Aircraft.Fins.MASS_SCALER] = \
                0.32 * area_exp * num_fins * \
                (taper_ratio + 0.5) * gross_weight_exp / GRAV_ENGLISH_LBM
            J[Aircraft.Fins.MASS, Mission.Design.GROSS_MASS] = (
                (0.096 * area_exp * num_fins * scalar * (taper_ratio + 0.5))
                / gross_weight**0.7)



"""
Contains any preliminary calculations on the fuselage.
"""

import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class FuselagePrelim(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuselage.LENGTH, val=0.0)
        add_aviary_input(self, Aircraft.Fuselage.MAX_HEIGHT, val=0.0)
        add_aviary_input(self, Aircraft.Fuselage.MAX_WIDTH, val=0.0)

        add_aviary_output(self, Aircraft.Fuselage.AVG_DIAMETER, val=0.0)
        add_aviary_output(self, Aircraft.Fuselage.PLANFORM_AREA, val=0.0)

    def setup_partials(self):
        self.declare_partials(of=[Aircraft.Fuselage.AVG_DIAMETER],
                              wrt=[Aircraft.Fuselage.MAX_HEIGHT,
                                   Aircraft.Fuselage.MAX_WIDTH],
                              val=0.5)
        self.declare_partials(of=[Aircraft.Fuselage.PLANFORM_AREA],
                              wrt=[Aircraft.Fuselage.LENGTH,
                                   Aircraft.Fuselage.MAX_WIDTH])

    def compute(self, inputs, outputs):
        max_height = inputs[Aircraft.Fuselage.MAX_HEIGHT]
        max_width = inputs[Aircraft.Fuselage.MAX_WIDTH]
        length = inputs[Aircraft.Fuselage.LENGTH]

        avg_diameter = 0.5 * (max_height + max_width)
        outputs[Aircraft.Fuselage.AVG_DIAMETER] = avg_diameter

        outputs[Aircraft.Fuselage.PLANFORM_AREA] = length * max_width

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        max_width = inputs[Aircraft.Fuselage.MAX_WIDTH]
        length = inputs[Aircraft.Fuselage.LENGTH]

        partials[Aircraft.Fuselage.PLANFORM_AREA,
                 Aircraft.Fuselage.LENGTH] = max_width

        partials[Aircraft.Fuselage.PLANFORM_AREA,
                 Aircraft.Fuselage.MAX_WIDTH] = length



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials
from parameterized import parameterized

from aviary.subsystems.geometry.flops_based.canard import Canard
from aviary.subsystems.geometry.flops_based.characteristic_lengths import \
    CharacteristicLengths
from aviary.subsystems.geometry.flops_based.fuselage import FuselagePrelim
from aviary.subsystems.geometry.flops_based.nacelle import Nacelles
from aviary.subsystems.geometry.flops_based.prep_geom import (PrepGeom,
                                                              _Fuselage,
                                                              _Prelim, _Tail,
                                                              _Wing)
from aviary.subsystems.geometry.flops_based.utils import Names
from aviary.subsystems.geometry.flops_based.wing import WingPrelim
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (do_validation_test,
                                                      flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_data,
                                                      get_flops_inputs,
                                                      get_flops_outputs,
                                                      print_case)
from aviary.variable_info.functions import override_aviary_vars
from aviary.variable_info.variables import Aircraft

unit_data_sets = get_flops_case_names(
    only=['LargeSingleAisle2FLOPS', 'LargeSingleAisle2FLOPSdw', 'LargeSingleAisle2FLOPSalt', 'LargeSingleAisle1FLOPS'])
wetted_area_overide = get_flops_case_names(
    only=['N3CC', 'LargeSingleAisle2FLOPS', 'LargeSingleAisle2FLOPSdw', 'LargeSingleAisle2FLOPSalt', 'LargeSingleAisle1FLOPS'])


# TODO: We have no integration tests for canard, so canard-related names are commented
# out.
class PrepGeomTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):
        class Statics(om.Group):

            def initialize(self):
                self.options.declare(
                    'aviary_options', types=AviaryValues,
                    desc='collection of Aircraft/Mission specific options')

            def setup(self):
                aviary_options = self.options['aviary_options']

                self.add_subsystem('prep_geom', PrepGeom(aviary_options=aviary_options),
                                   promotes=['*'])

            def configure(self):
                aviary_options = self.options['aviary_options']

                override_aviary_vars(self, aviary_options)

        options = get_flops_data(case_name,
                                 keys=[
                                     Aircraft.Fuselage.NUM_FUSELAGES,
                                     Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES,
                                     Aircraft.VerticalTail.NUM_TAILS,
                                     Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION,
                                     Aircraft.Engine.NUM_ENGINES,
                                     Aircraft.Propulsion.TOTAL_NUM_ENGINES,
                                     'engine_models'
                                 ])

        prob = self.prob

        prob.model.add_subsystem(
            'statics', Statics(aviary_options=options), promotes=['*'])

        prob.setup(check=False, force_alloc_complex=True)

        output_keys = [Aircraft.Fuselage.AVG_DIAMETER,
                       Aircraft.Fuselage.CHARACTERISTIC_LENGTH,
                       Aircraft.Fuselage.CROSS_SECTION,
                       Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
                       Aircraft.Fuselage.FINENESS,
                       Aircraft.Fuselage.LENGTH_TO_DIAMETER,

                       # Aircraft.Canard.CHARACTERISTIC_LENGTH,
                       # Aircraft.Canard.FINENESS,

                       Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH,
                       Aircraft.HorizontalTail.FINENESS,

                       Aircraft.Nacelle.CHARACTERISTIC_LENGTH,
                       Aircraft.Nacelle.FINENESS,

                       Aircraft.VerticalTail.CHARACTERISTIC_LENGTH,
                       Aircraft.VerticalTail.FINENESS,

                       Aircraft.Wing.CHARACTERISTIC_LENGTH,
                       Aircraft.Wing.FINENESS]

        if case_name not in wetted_area_overide:
            output_keys.extend([Aircraft.Canard.WETTED_AREA,
                                Aircraft.Design.TOTAL_WETTED_AREA,
                                Aircraft.Fuselage.WETTED_AREA,
                                Aircraft.HorizontalTail.WETTED_AREA,
                                Aircraft.Nacelle.TOTAL_WETTED_AREA,
                                Aircraft.Nacelle.WETTED_AREA,
                                Aircraft.VerticalTail.WETTED_AREA,
                                Aircraft.Wing.WETTED_AREA])

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Fuselage.LENGTH,
                        Aircraft.Fuselage.MAX_HEIGHT,
                        Aircraft.Fuselage.MAX_WIDTH,
                        Aircraft.Fuselage.WETTED_AREA_SCALER,

                        # Aircraft.Canard.AREA,
                        # Aircraft.Canard.ASPECT_RATIO,
                        # Aircraft.Canard.THICKNESS_TO_CHORD,
                        # Aircraft.Canard.WETTED_AREA_SCALER,

                        Aircraft.HorizontalTail.AREA,
                        Aircraft.HorizontalTail.ASPECT_RATIO,
                        Aircraft.HorizontalTail.TAPER_RATIO,
                        Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
                        Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION,
                        Aircraft.HorizontalTail.WETTED_AREA_SCALER,

                        Aircraft.Nacelle.AVG_DIAMETER,
                        Aircraft.Nacelle.AVG_LENGTH,
                        # Aircraft.Nacelle.WETTED_AREA_SCALER,

                        Aircraft.VerticalTail.AREA,
                        Aircraft.VerticalTail.ASPECT_RATIO,
                        Aircraft.VerticalTail.TAPER_RATIO,
                        Aircraft.VerticalTail.THICKNESS_TO_CHORD,
                        Aircraft.VerticalTail.WETTED_AREA_SCALER,

                        Aircraft.Wing.AREA,
                        Aircraft.Wing.ASPECT_RATIO,
                        Aircraft.Wing.GLOVE_AND_BAT,
                        Aircraft.Wing.SPAN,
                        Aircraft.Wing.TAPER_RATIO,
                        Aircraft.Wing.THICKNESS_TO_CHORD,
                        Aircraft.Wing.WETTED_AREA_SCALER],
            output_keys=output_keys,
            aviary_option_keys=[Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION,
                                Aircraft.Fuselage.NUM_FUSELAGES],
            tol=1e-2,
            atol=1e-6,
            rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)

    def test_prelim_fuselage(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', FuselagePrelim(),
                            promotes=['*'])

        prob.setup(force_alloc_complex=True)

        prob.set_val(Aircraft.Fuselage.LENGTH, 22.0)
        prob.set_val(Aircraft.Fuselage.MAX_HEIGHT, 7.0)
        prob.set_val(Aircraft.Fuselage.MAX_WIDTH, 4.0)

        partial_data = prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)

    def test_prelim_wing(self):
        prob = om.Problem()
        model = prob.model

        model.add_subsystem('comp', WingPrelim(),
                            promotes=['*'])

        prob.setup(force_alloc_complex=True)

        prob.set_val(Aircraft.Wing.AREA, 22.0)
        prob.set_val(Aircraft.Wing.GLOVE_AND_BAT, 7.0)
        prob.set_val(Aircraft.Wing.SPAN, 4.0)

        partial_data = prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-12, rtol=1e-12)


class _PrelimTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(unit_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'prelim',
            _Prelim(aviary_options=get_flops_inputs(case_name,
                                                    keys=[Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION])),
            promotes=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=get_flops_data(case_name),
                           output_validation_data=local_variables[case_name],
                           input_keys=[Aircraft.Fuselage.AVG_DIAMETER,
                                       Aircraft.Fuselage.MAX_WIDTH,
                                       Aircraft.HorizontalTail.AREA,
                                       Aircraft.HorizontalTail.ASPECT_RATIO,
                                       Aircraft.HorizontalTail.TAPER_RATIO,
                                       Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
                                       Aircraft.VerticalTail.AREA,
                                       Aircraft.VerticalTail.ASPECT_RATIO,
                                       Aircraft.VerticalTail.TAPER_RATIO,
                                       Aircraft.VerticalTail.THICKNESS_TO_CHORD,
                                       Aircraft.Wing.AREA,
                                       Aircraft.Wing.GLOVE_AND_BAT,
                                       Aircraft.Wing.SPAN,
                                       Aircraft.Wing.TAPER_RATIO,
                                       Aircraft.Wing.THICKNESS_TO_CHORD],
                           output_keys=[Names.CROOT,
                                        Names.CROOTB,
                                        Names.CROTM,
                                        Names.CROTVT,
                                        Names.CRTHTB,
                                        Names.SPANHT,
                                        Names.SPANVT,
                                        Names.XDX,
                                        Names.XMULT,
                                        Names.XMULTH,
                                        Names.XMULTV],
                           aviary_option_keys=Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION,
                           tol=1e-3,
                           atol=1e-6,
                           rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class _WingTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(unit_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'wings',
            _Wing(aviary_options=get_flops_inputs(case_name,
                                                  keys=[Aircraft.Fuselage.NUM_FUSELAGES])),
            promotes=['*'])

        prob.setup(check=False, force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=_hybrid_input_data(case_name),
                           output_validation_data=get_flops_outputs(case_name),
                           input_keys=[Names.CROOT,
                                       Names.CROOTB,
                                       Names.XDX,
                                       Names.XMULT,
                                       Aircraft.Wing.AREA,
                                       Aircraft.Wing.WETTED_AREA_SCALER],
                           output_keys=Aircraft.Wing.WETTED_AREA,
                           aviary_option_keys=Aircraft.Fuselage.NUM_FUSELAGES,
                           check_values=case_name not in wetted_area_overide,
                           tol=1e-3,
                           atol=1e-6,
                           rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class _TailTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(unit_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'tails',
            _Tail(aviary_options=get_flops_inputs(case_name,
                                                  keys=[Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION,
                                                        Aircraft.Propulsion.TOTAL_NUM_FUSELAGE_ENGINES])),
            promotes=['*'])

        prob.setup(check=False, force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=_hybrid_input_data(case_name),
                           output_validation_data=get_flops_outputs(case_name),
                           input_keys=[Names.XMULTH,
                                       Names.XMULTV,
                                       Aircraft.HorizontalTail.AREA,
                                       Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION,
                                       Aircraft.HorizontalTail.WETTED_AREA_SCALER,
                                       Aircraft.VerticalTail.AREA,
                                       Aircraft.VerticalTail.WETTED_AREA_SCALER],
                           output_keys=[Aircraft.HorizontalTail.WETTED_AREA,
                                        Aircraft.VerticalTail.WETTED_AREA],
                           aviary_option_keys=Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION,
                           check_values=case_name not in wetted_area_overide,
                           tol=1e-3,
                           atol=1e-6,
                           rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class _FuselageTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(unit_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'fuse',
            _Fuselage(aviary_options=get_flops_inputs(case_name,
                                                      keys=[Aircraft.Fuselage.NUM_FUSELAGES])),
            promotes=['*'])

        prob.setup(check=False, force_alloc_complex=True)

        outputs = [
            Aircraft.Fuselage.CROSS_SECTION,
            Aircraft.Fuselage.DIAMETER_TO_WING_SPAN,
            Aircraft.Fuselage.LENGTH_TO_DIAMETER
        ]

        if case_name not in wetted_area_overide:
            outputs.extend(Aircraft.Fuselage.WETTED_AREA)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=_hybrid_input_data(case_name),
                           output_validation_data=get_flops_outputs(case_name),
                           input_keys=[Names.CROOTB,
                                       Names.CROTVT,
                                       Names.CRTHTB,
                                       Aircraft.Fuselage.AVG_DIAMETER,
                                       Aircraft.Fuselage.LENGTH,
                                       Aircraft.Fuselage.WETTED_AREA_SCALER,
                                       Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
                                       Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION,
                                       Aircraft.VerticalTail.THICKNESS_TO_CHORD,
                                       Aircraft.Wing.AREA,
                                       Aircraft.Wing.ASPECT_RATIO,
                                       Aircraft.Wing.GLOVE_AND_BAT,
                                       Aircraft.Wing.THICKNESS_TO_CHORD],
                           output_keys=[Aircraft.HorizontalTail.WETTED_AREA,
                                        Aircraft.VerticalTail.WETTED_AREA],
                           aviary_option_keys=Aircraft.Fuselage.NUM_FUSELAGES,
                           check_values=case_name not in wetted_area_overide,
                           tol=1e-2,
                           atol=1e-6,
                           rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class NacellesTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(unit_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'nacelles',
            Nacelles(aviary_options=get_flops_inputs(case_name,
                                                     keys=[Aircraft.Engine.NUM_ENGINES])),
            promotes=['*'])

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(prob,
                              case_name,
                              input_keys=[Aircraft.Nacelle.AVG_DIAMETER,
                                          Aircraft.Nacelle.AVG_LENGTH,
                                          Aircraft.Nacelle.WETTED_AREA_SCALER],
                              output_keys=[Aircraft.Nacelle.TOTAL_WETTED_AREA,
                                           Aircraft.Nacelle.WETTED_AREA],
                              aviary_option_keys=[Aircraft.Engine.NUM_ENGINES],
                              tol=1e-3,
                              atol=1e-6,
                              rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class CanardTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    def test_case(self):
        prob = self.prob

        prob.model.add_subsystem(
            'canard',
            Canard(aviary_options=AviaryValues()),
            promotes=['*'])

        prob.setup(check=False, force_alloc_complex=True)

        do_validation_test(prob,
                           'canard_test',
                           input_validation_data=Canard_test_data,
                           output_validation_data=Canard_test_data,
                           input_keys=[Aircraft.Canard.AREA,
                                       Aircraft.Canard.THICKNESS_TO_CHORD,
                                       Aircraft.Canard.WETTED_AREA_SCALER],
                           output_keys=Aircraft.Canard.WETTED_AREA,
                           aviary_option_keys=[Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION,
                                               Aircraft.Fuselage.NUM_FUSELAGES],
                           tol=1e-2,
                           atol=1e-6,
                           rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class CharacteristicLengthsTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(unit_data_sets,
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'characteristic_lengths',
            CharacteristicLengths(aviary_options=get_flops_inputs(case_name,
                                                                  keys=[Aircraft.Engine.NUM_ENGINES,
                                                                        Aircraft.Fuselage.NUM_FUSELAGES,
                                                                        Aircraft.VerticalTail.NUM_TAILS,
                                                                        Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION])),
            promotes=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        do_validation_test(prob,
                           case_name,
                           input_validation_data=_hybrid_input_data(case_name),
                           output_validation_data=get_flops_outputs(case_name),
                           input_keys=[Names.CROOT,
                                       Aircraft.Canard.AREA,
                                       Aircraft.Canard.ASPECT_RATIO,
                                       Aircraft.Canard.THICKNESS_TO_CHORD,
                                       Aircraft.Fuselage.AVG_DIAMETER,
                                       Aircraft.Fuselage.LENGTH,
                                       Aircraft.HorizontalTail.AREA,
                                       Aircraft.HorizontalTail.ASPECT_RATIO,
                                       Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
                                       Aircraft.Nacelle.AVG_DIAMETER,
                                       Aircraft.Nacelle.AVG_LENGTH,
                                       Aircraft.VerticalTail.AREA,
                                       Aircraft.VerticalTail.ASPECT_RATIO,
                                       Aircraft.VerticalTail.THICKNESS_TO_CHORD,
                                       Aircraft.Wing.AREA,
                                       Aircraft.Wing.ASPECT_RATIO,
                                       Aircraft.Wing.GLOVE_AND_BAT,
                                       Aircraft.Wing.TAPER_RATIO,
                                       Aircraft.Wing.THICKNESS_TO_CHORD],
                           output_keys=[Aircraft.Canard.CHARACTERISTIC_LENGTH,
                                        Aircraft.Canard.FINENESS,
                                        Aircraft.Fuselage.CHARACTERISTIC_LENGTH,
                                        Aircraft.Fuselage.FINENESS,
                                        Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH,
                                        Aircraft.HorizontalTail.FINENESS,
                                        Aircraft.Nacelle.CHARACTERISTIC_LENGTH,
                                        Aircraft.Nacelle.FINENESS,
                                        Aircraft.VerticalTail.CHARACTERISTIC_LENGTH,
                                        Aircraft.VerticalTail.FINENESS,
                                        Aircraft.Wing.CHARACTERISTIC_LENGTH,
                                        Aircraft.Wing.FINENESS],
                           aviary_option_keys=[
                               Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION, Aircraft.Propulsion.TOTAL_NUM_ENGINES],
                           tol=1e-2,
                           atol=1e-6,
                           rtol=1e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


def _hybrid_input_data(case_name):
    data = get_flops_inputs(case_name)
    data.update(local_variables[case_name])
    return data


_LargeSingleAisle_custom_test_data = AviaryValues({
    Names.CROOT: (16.415788, 'unitless'),
    Names.CROOTB: (15.133300, 'unitless'),
    Names.CROTM: (0.9218747347874056, 'unitless'),
    Names.CROTVT: (19.156603, 'unitless'),
    Names.CRTHTB: (13.502073, 'unitless'),
    Names.SPANHT: (46.151923036857305, 'unitless'),
    Names.SPANVT: (22.293496809607955, 'unitless'),
    Names.XDX: (12.750000, 'unitless'),
    Names.XMULT: (2.050310, 'unitless'),
    Names.XMULTH: (2.048375, 'unitless'),
    Names.XMULTV: (2.046247, 'unitless'),
})

_other_custom_test_data = AviaryValues({
    Names.CROOT: (19.254769, 'unitless'),
    Names.CROOTB: (17.601290, 'unitless'),
    Names.CROTM: (0.914126224442001, 'unitless'),
    Names.CROTVT: (18.672780, 'unitless'),
    Names.CRTHTB: (14.205434, 'unitless'),
    Names.SPANHT: (47.09069715015741, 'unitless'),
    Names.SPANVT: (25.166513637040453, 'unitless'),
    Names.XDX: (12.675400, 'unitless'),
    Names.XMULT: (2.050981, 'unitless'),
    Names.XMULTH: (2.046247, 'unitless'),
    Names.XMULTV: (2.053197, 'unitless'),
})

local_variables = {}
local_variables['LargeSingleAisle1FLOPS'] = _LargeSingleAisle_custom_test_data
local_variables['LargeSingleAisle2FLOPS'] = _other_custom_test_data
local_variables['LargeSingleAisle2FLOPSdw'] = _other_custom_test_data
local_variables['LargeSingleAisle2FLOPSalt'] = _other_custom_test_data

# NOTE: no current data set includes canard
Canard_test_data = AviaryValues({
    Aircraft.Canard.AREA: (10.0, 'ft**2'),
    Aircraft.Canard.ASPECT_RATIO: (1.5, 'unitless'),
    Aircraft.Canard.CHARACTERISTIC_LENGTH: (2.58, 'ft'),
    Aircraft.Canard.FINENESS: (0.3000, 'unitless'),
    Aircraft.Canard.TAPER_RATIO: (0.5, 'unitless'),
    Aircraft.Canard.THICKNESS_TO_CHORD: (0.3, 'unitless'),
    Aircraft.Canard.WETTED_AREA: (21.16, 'ft**2'),
    Aircraft.Canard.WETTED_AREA_SCALER: (1.0, 'unitless'),
})


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.starter import TransportStarterMass
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission


class TransportStarterMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(omit='N3CC'),
                          name_func=print_case)
    def test_case_1(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            "starter_test",
            TransportStarterMass(aviary_options=get_flops_inputs(case_name)),
            promotes_outputs=['*'],
            promotes_inputs=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Nacelle.AVG_DIAMETER],
            output_keys=Aircraft.Propulsion.TOTAL_STARTER_MASS)

    def test_case_2(self):
        # test with more than 4 engines
        prob = self.prob

        aviary_options = get_flops_inputs('LargeSingleAisle1FLOPS')
        aviary_options.set_val(Aircraft.Engine.NUM_ENGINES, np.array([5]))
        aviary_options.set_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES, 5)
        aviary_options.set_val(Mission.Constraints.MAX_MACH, 0.785)

        prob.model.add_subsystem(
            "starter_test",
            TransportStarterMass(aviary_options=aviary_options),
            promotes_outputs=['*'],
            promotes_inputs=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        prob.set_val(Aircraft.Nacelle.AVG_DIAMETER, np.array([7.94]), 'ft')

        prob.run_model()

        mass = prob.get_val(Aircraft.Propulsion.TOTAL_STARTER_MASS, 'lbm')
        expected_mass = 1555.38298314

        assert_near_equal(mass, expected_mass, tolerance=1e-10)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-10)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



import openmdao.api as om

from aviary.variable_info.variables import Aircraft, Mission


class ParamPort(om.ExplicitComponent):

    param_data = {
        Aircraft.Wing.AREA: dict(units="ft**2", val=1370.3),
        Aircraft.Wing.INCIDENCE: dict(units="deg", val=0),
        Aircraft.Wing.HEIGHT: dict(units="ft", val=8),
        Aircraft.Wing.SPAN: dict(units="ft", val=117.8),
        Mission.Design.GROSS_MASS: dict(units="lbm", val=175400),
        Mission.Summary.GROSS_MASS: dict(units="lbm", val=175400),
        Mission.Summary.FUEL_FLOW_SCALER: dict(units="unitless", val=1.0),
        Mission.Takeoff.AIRPORT_ALTITUDE: dict(units="ft", val=0),
        Mission.Landing.AIRPORT_ALTITUDE: dict(units="ft", val=0),
        Aircraft.Wing.FLAP_DEFLECTION_TAKEOFF: dict(units="deg", val=10),
        Aircraft.Wing.FLAP_DEFLECTION_LANDING: dict(units="deg", val=40),
        Aircraft.Wing.AVERAGE_CHORD: dict(units="ft", val=12.615),
        Aircraft.Fuselage.AVG_DIAMETER: dict(units="inch", val=12 * 13.100),
        Aircraft.HorizontalTail.AVERAGE_CHORD: dict(units="ft", val=9.577),
        Aircraft.HorizontalTail.AREA: dict(units="ft**2", val=375.880),
        Aircraft.HorizontalTail.SPAN: dict(units="ft", val=42.254),
        Aircraft.VerticalTail.AVERAGE_CHORD: dict(units="ft", val=16.832),
        Aircraft.VerticalTail.AREA: dict(units="ft**2", val=469.318),
        Aircraft.VerticalTail.SPAN: dict(units="ft", val=27.996),
        Aircraft.Fuselage.LENGTH: dict(units="ft", val=129.4),
        Aircraft.Nacelle.AVG_LENGTH: dict(units="ft", val=14.5),
        Aircraft.Fuselage.WETTED_AREA: dict(units="ft**2", val=4000),
        Aircraft.Nacelle.SURFACE_AREA: dict(units="ft**2", val=659.23 / 2),
        Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED: dict(units="unitless", val=0.1397),
        Aircraft.Strut.CHORD: dict(
            units="ft", val=0
        ),  # only available if Aviary_option Aircraft.Wing.HAS_STRUT
        Aircraft.Wing.ASPECT_RATIO: dict(units="unitless", val=10.13),
        Aircraft.Wing.TAPER_RATIO: dict(units="unitless", val=0.33),
        Aircraft.Wing.THICKNESS_TO_CHORD_ROOT: dict(units="unitless", val=0.15),
        Aircraft.Wing.THICKNESS_TO_CHORD_TIP: dict(units="unitless", val=0.12),
        Aircraft.HorizontalTail.VERTICAL_TAIL_FRACTION: dict(units="unitless", val=0),
        Aircraft.Wing.SWEEP: dict(units="deg", val=25),
        Aircraft.HorizontalTail.SWEEP: dict(units="deg", val=25),
        Aircraft.HorizontalTail.MOMENT_RATIO: dict(units="unitless", val=0.2307),
        Aircraft.Wing.MOUNTING_TYPE: dict(units="unitless", val=0),
        Aircraft.Design.STATIC_MARGIN: dict(units="unitless", val=0.03),
        Aircraft.Design.CG_DELTA: dict(units="unitless", val=0.25),
        Aircraft.Wing.FORM_FACTOR: dict(units="unitless", val=1.25),
        Aircraft.Fuselage.FORM_FACTOR: dict(units="unitless", val=1.25),
        Aircraft.Nacelle.FORM_FACTOR: dict(units="unitless", val=1.5),
        Aircraft.VerticalTail.FORM_FACTOR: dict(units="unitless", val=1.25),
        Aircraft.HorizontalTail.FORM_FACTOR: dict(units="unitless", val=1.25),
        Aircraft.Wing.FUSELAGE_INTERFERENCE_FACTOR: dict(units="unitless", val=1.1),
        Aircraft.Strut.FUSELAGE_INTERFERENCE_FACTOR: dict(units="unitless", val=0),
        Aircraft.Design.DRAG_COEFFICIENT_INCREMENT: dict(units="unitless", val=0.00175),
        Aircraft.Fuselage.FLAT_PLATE_AREA_INCREMENT: dict(units="ft**2", val=0.25),
        Aircraft.Wing.CENTER_DISTANCE: dict(units="unitless", val=0.463),
        Aircraft.Wing.MIN_PRESSURE_LOCATION: dict(units="unitless", val=0.3),
        Aircraft.Wing.MAX_THICKNESS_LOCATION: dict(units="unitless", val=0.4),
        Aircraft.Strut.AREA_RATIO: dict(units="unitless", val=0),
        Aircraft.Wing.ZERO_LIFT_ANGLE: dict(units="deg", val=-1.2),
        Aircraft.Design.SUPERCRITICAL_DIVERGENCE_SHIFT: dict(units="unitless", val=0.033),
        Aircraft.Wing.FLAP_CHORD_RATIO: dict(units="unitless", val=0.3),
        Mission.Design.LIFT_COEFFICIENT_MAX_FLAPS_UP: dict(units="unitless", val=1.2596),
        Mission.Takeoff.LIFT_COEFFICIENT_MAX: dict(units="unitless", val=2.1886),
        Mission.Landing.LIFT_COEFFICIENT_MAX: dict(units="unitless", val=2.8155),
        Mission.Takeoff.LIFT_COEFFICIENT_FLAP_INCREMENT: dict(
            units="unitless", val=0.4182
        ),
        Mission.Landing.LIFT_COEFFICIENT_FLAP_INCREMENT: dict(
            units="unitless", val=1.0293
        ),
        Mission.Takeoff.DRAG_COEFFICIENT_FLAP_INCREMENT: dict(
            units="unitless", val=0.0085
        ),
        Mission.Landing.DRAG_COEFFICIENT_FLAP_INCREMENT: dict(
            units="unitless", val=0.0406
        ),
    }

    def setup(self):
        # override
        for name, data in self.param_data.items():
            self.add_input(name, **data)

    @staticmethod
    def set_default_vals(sys):
        """Set input defaults on a group"""
        for name, data in ParamPort.param_data.items():
            sys.set_input_defaults(
                name, val=data.get("val", 1), units=data.get("units", None)
            )

    @staticmethod
    def add_params(param_data):
        """Overlay additional parameters for specialized dynamic systems.

        Parameters
        ----------
        param_data : dict
            Dictionary with parameter names as keys and sub-dictionaries for values,
            containing optionally ``val`` and ``units`` keys to set default value and
            units for the parameter.
        """
        ParamPort.param_data.update(param_data)

    @staticmethod
    def promote_params(sys, trajs=None, phases=None):
        """Add parameters to trajectories and/or phases and set defaults.

        Parameters are also promoted at the ``sys`` level.

        If trajectories are provided, the phases for each trajectory must also be
        supplied. If only phases are provided, no trajectory-level parameters are added.

        Parameters
        ----------
        sys : System
            System of which the trajectories or phases are subsystems.
        trajs : iterable, optional
            Trajectories to which to add parameters. If ``None`` (default), only phase
            parameters are added.
        phases : iterable, optional
            Phases to which to add parameters.
        """
        proms = [(f"parameters:{param}", param) for param in ParamPort.param_data]

        if trajs:
            for trajname, phasenames in zip(trajs, phases):
                traj = getattr(sys, trajname)
                for param, data in ParamPort.param_data.items():
                    traj.add_parameter(
                        param,
                        units=data.get("units", None),
                        val=data.get("val", 1.0),
                        opt=False,
                        static_target=True,
                        targets={phasename: [param] for phasename in phasenames},
                    )
                sys.promotes(trajname, inputs=proms)
        else:
            for phasename in phases:
                phase = getattr(sys, phasename)
                for param, data in ParamPort.param_data.items():
                    phase.add_parameter(
                        param,
                        units=data.get("units", None),
                        val=data.get("val", 1.0),
                        opt=False,
                        static_target=True,
                    )
                sys.promotes(phasename, inputs=proms)



import numpy as np
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.variable_info.enums import AlphaModes, AnalysisScheme
from aviary.variable_info.variables import Aircraft, Mission, Dynamic
from aviary.mission.gasp_based.ode.ascent_eom import AscentEOM

from aviary.mission.gasp_based.ode.base_ode import BaseODE
from aviary.mission.gasp_based.ode.params import ParamPort
from aviary.variable_info.enums import AlphaModes
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class AscentODE(BaseODE):
    """ODE for initial ascent.

    This phase is intended to model the portion of aircraft flight starting when the
    aircraft detaches from the runway, and it is capable of retracting flaps and landing
    gear as necessary.
    """

    def initialize(self):
        super().initialize()
        self.options.declare("alpha_mode", default=AlphaModes.DEFAULT, types=AlphaModes)

    def setup(self):
        nn = self.options["num_nodes"]
        alpha_mode = self.options["alpha_mode"]
        analysis_scheme = self.options["analysis_scheme"]
        aviary_options = self.options['aviary_options']
        core_subsystems = self.options['core_subsystems']

        # TODO: paramport
        ascent_params = ParamPort()
        if analysis_scheme is AnalysisScheme.SHOOTING:
            ascent_params.add_params({
                Aircraft.Design.MAX_FUSELAGE_PITCH_ANGLE: dict(units='deg', val=0),
            })
        self.add_subsystem("params", ascent_params, promotes=["*"])

        self.add_subsystem(
            "USatm",
            USatm1976Comp(
                num_nodes=nn),
            promotes_inputs=[
                ("h",
                 Dynamic.Mission.ALTITUDE)],
            promotes_outputs=[
                "rho",
                ("sos",
                 Dynamic.Mission.SPEED_OF_SOUND),
                ("temp",
                 Dynamic.Mission.TEMPERATURE),
                ("pres",
                 Dynamic.Mission.STATIC_PRESSURE),
                "viscosity"],
        )

        self.add_flight_conditions(nn)

        kwargs = {'num_nodes': nn, 'aviary_inputs': aviary_options,
                  'method': 'low_speed', 'retract_gear': True, 'retract_flaps': True}
        for subsystem in core_subsystems:
            system = subsystem.build_mission(**kwargs)
            if system is not None:
                self.add_subsystem(subsystem.name,
                                   system,
                                   promotes_inputs=subsystem.mission_inputs(**kwargs),
                                   promotes_outputs=subsystem.mission_outputs(**kwargs))

        if alpha_mode is AlphaModes.DEFAULT:
            # alpha as input
            pass
        else:
            self.AddAlphaControl(alpha_mode=alpha_mode, num_nodes=nn)

        if analysis_scheme is AnalysisScheme.SHOOTING:
            shooting_inputs = [Dynamic.Mission.DISTANCE]
        else:
            shooting_inputs = []

        self.add_subsystem(
            "eoms",
            AscentEOM(num_nodes=nn,
                      analysis_scheme=analysis_scheme),
            promotes_inputs=[
                Dynamic.Mission.MASS,
                Dynamic.Mission.THRUST_TOTAL,
                Dynamic.Mission.LIFT,
                Dynamic.Mission.DRAG,
                Dynamic.Mission.VELOCITY,
                Dynamic.Mission.FLIGHT_PATH_ANGLE,
                "alpha",
            ]
            + shooting_inputs
            + ["aircraft:*"],
            promotes_outputs=[
                Dynamic.Mission.VELOCITY_RATE,
                Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE,
                Dynamic.Mission.ALTITUDE_RATE,
                Dynamic.Mission.DISTANCE_RATE,
                "alpha_rate",
                "normal_force",
                "fuselage_pitch",
                "load_factor",
            ],
        )

        self.add_excess_rate_comps(nn)

        ParamPort.set_default_vals(self)
        self.set_input_defaults("t_init_flaps", val=47.5)
        self.set_input_defaults("t_init_gear", val=37.3)
        self.set_input_defaults("alpha", val=np.zeros(nn), units="deg")
        self.set_input_defaults(Dynamic.Mission.FLIGHT_PATH_ANGLE,
                                val=np.zeros(nn), units="deg")
        self.set_input_defaults(Dynamic.Mission.ALTITUDE, val=np.zeros(nn), units="ft")
        self.set_input_defaults(Dynamic.Mission.VELOCITY, val=np.zeros(nn), units="kn")
        self.set_input_defaults("t_curr", val=np.zeros(nn), units="s")
        self.set_input_defaults('aero_ramps.flap_factor:final_val', val=0.)
        self.set_input_defaults('aero_ramps.gear_factor:final_val', val=0.)
        self.set_input_defaults('aero_ramps.flap_factor:initial_val', val=1.)
        self.set_input_defaults('aero_ramps.gear_factor:initial_val', val=1.)
        self.set_input_defaults(Dynamic.Mission.MASS, val=np.ones(
            nn), units='kg')  # val here is nominal



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class TransportCargoContainersMass(om.ExplicitComponent):
    '''
    Calculate the estimated mass for cargo containers. Use for both
    traditional and blended-wing-body type transports. The methodology is based on
    the FLOPS weight equations, modified to output mass instead of weight.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(
            self, Aircraft.CrewPayload.CARGO_CONTAINER_MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.CrewPayload.CARGO_MASS, val=0.0)

        add_aviary_input(self, Aircraft.CrewPayload.BAGGAGE_MASS, val=0.0)

        add_aviary_output(self, Aircraft.CrewPayload.CARGO_CONTAINER_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        scaler = inputs[Aircraft.CrewPayload.CARGO_CONTAINER_MASS_SCALER]
        cargo = inputs[Aircraft.CrewPayload.CARGO_MASS]
        baggage = inputs[Aircraft.CrewPayload.BAGGAGE_MASS]

        temp = (cargo + baggage) / 950.0 + 0.99
        container_count = sin_int4(temp)
        cargo_container_weight = container_count * 175.0 * scaler

        outputs[Aircraft.CrewPayload.CARGO_CONTAINER_MASS] = \
            cargo_container_weight / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J, discrete_inputs=None):
        scaler = inputs[Aircraft.CrewPayload.CARGO_CONTAINER_MASS_SCALER]
        cargo = inputs[Aircraft.CrewPayload.CARGO_MASS]
        baggage = inputs[Aircraft.CrewPayload.BAGGAGE_MASS]

        temp = (cargo + baggage) / 950.0 + 0.99
        container_count = sin_int4(temp)

        partial = dydx_sin_int4(temp) / 950.0 * 175.0

        J[
            Aircraft.CrewPayload.CARGO_CONTAINER_MASS,
            Aircraft.CrewPayload.CARGO_CONTAINER_MASS_SCALER
        ] = container_count * 175.0 / GRAV_ENGLISH_LBM

        J[
            Aircraft.CrewPayload.CARGO_CONTAINER_MASS,
            Aircraft.CrewPayload.CARGO_MASS
        ] = partial * scaler / GRAV_ENGLISH_LBM

        J[
            Aircraft.CrewPayload.CARGO_CONTAINER_MASS,
            Aircraft.CrewPayload.BAGGAGE_MASS
        ] = partial * scaler / GRAV_ENGLISH_LBM


# region TODO: move this to an appropriate module for import
def sin_int4(val):
    '''
    Define a smooth, differentialbe approximation to the 'int' function.
    '''
    return sin_int(sin_int(sin_int(sin_int(val)))) - 0.5


def dydx_sin_int4(val):
    '''
    Define the derivative (dy/dx) of sin_int4, at x = val.
    '''
    y0 = sin_int(val)
    y1 = sin_int(y0)
    y2 = sin_int(y1)

    dydx3 = dydx_sin_int(y2)
    dydx2 = dydx_sin_int(y1)
    dydx1 = dydx_sin_int(y0)
    dydx0 = dydx_sin_int(val)

    dydx = dydx3 * dydx2 * dydx1 * dydx0

    return dydx


# 'int' function can be approximated by recursively applying this sin function
# which makes a smooth, differentialbe function
def sin_int(val):
    '''
    Define one step in approximating the 'int' function with a smooth,
    differentialbe function.
    '''
    int_val = val - np.sin(2*np.pi*(val+0.5))/(2*np.pi)

    return int_val


def dydx_sin_int(val):
    '''
    Define the derivative (dy/dx) of sin_int, at x = val.
    '''
    dydx = 1.0 - np.cos(2 * np.pi * (val + 0.5))

    return dydx
# endregion TODO: move this to an appropriate module for import



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft

# TODO should additional misc mass be separated out into a separate component?


class EngineMass(om.ExplicitComponent):
    '''
    Calculate scaled engine mass and additional miscellaneous mass for a
    single named engine model.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        count = len(self.options['aviary_options'].get_val('engine_models'))

        add_aviary_input(self, Aircraft.Engine.SCALED_SLS_THRUST, val=np.zeros(count))

        add_aviary_output(self, Aircraft.Engine.MASS, val=np.zeros(count))
        add_aviary_output(self, Aircraft.Engine.ADDITIONAL_MASS, val=np.zeros(count))
        add_aviary_output(self, Aircraft.Propulsion.TOTAL_ENGINE_MASS, val=0.0)

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']

        # cast to numpy arrays to ensure values are always correct type
        num_engines = np.array(aviary_options.get_val(Aircraft.Engine.NUM_ENGINES))
        scaling_parameter = np.array(aviary_options.get_val(Aircraft.Engine.MASS_SCALER))
        scale_mass = np.array(aviary_options.get_val(Aircraft.Engine.SCALE_MASS))
        addtl_mass_fraction = np.array(aviary_options.get_val(
            Aircraft.Engine.ADDITIONAL_MASS_FRACTION))
        ref_engine_mass = np.array(aviary_options.get_val(
            Aircraft.Engine.REFERENCE_MASS, units='lbm'))
        ref_thrust = np.array(aviary_options.get_val(
            Aircraft.Engine.REFERENCE_SLS_THRUST, units='lbf'))

        scaled_thrust = np.array(inputs[Aircraft.Engine.SCALED_SLS_THRUST])

        scale_idx = np.where(scale_mass)
        # indices where scaling is applied and scaling equation is used
        param_idx = np.where(scaling_parameter[scale_idx] >= 0.3)

        # default mass is reference mass
        # use dtype to make complex safe
        calc_mass = np.array(ref_engine_mass, dtype=scaled_thrust.dtype)

        # scale engine mass using equation chosen by value of user-provided mass scalar
        thrust_ratio = scaled_thrust / ref_thrust

        calc_mass[scale_idx] = ref_engine_mass[scale_idx] + (
            scaled_thrust[scale_idx] - ref_thrust[scale_idx]) * scaling_parameter[scale_idx]

        calc_mass[param_idx] = ref_engine_mass[param_idx] * \
            thrust_ratio[param_idx]**scaling_parameter[param_idx]

        addtl_mass = addtl_mass_fraction * calc_mass

        outputs[Aircraft.Engine.MASS] = calc_mass
        outputs[Aircraft.Propulsion.TOTAL_ENGINE_MASS] = sum(
            calc_mass * num_engines)
        outputs[Aircraft.Engine.ADDITIONAL_MASS] = addtl_mass

    def setup_partials(self):
        count = len(self.options['aviary_options'].get_val('engine_models'))
        shape = np.arange(count)

        self.declare_partials(Aircraft.Engine.MASS,
                              Aircraft.Engine.SCALED_SLS_THRUST,
                              rows=shape, cols=shape)
        self.declare_partials(
            Aircraft.Engine.ADDITIONAL_MASS,
            Aircraft.Engine.SCALED_SLS_THRUST,
            rows=shape, cols=shape)
        self.declare_partials(
            Aircraft.Propulsion.TOTAL_ENGINE_MASS,
            Aircraft.Engine.SCALED_SLS_THRUST)

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        engine_models = aviary_options.get_val('engine_models')
        count = len(engine_models)

        num_engines = np.array(aviary_options.get_val(Aircraft.Engine.NUM_ENGINES))
        scaling_parameter = np.array(aviary_options.get_val(Aircraft.Engine.MASS_SCALER))
        scale_mass = np.array(aviary_options.get_val(Aircraft.Engine.SCALE_MASS))
        addtl_mass_fraction = np.array(aviary_options.get_val(
            Aircraft.Engine.ADDITIONAL_MASS_FRACTION))
        ref_engine_mass = np.array(aviary_options.get_val(
            Aircraft.Engine.REFERENCE_MASS, units='lbm'))
        ref_sls_thrust = np.array(aviary_options.get_val(
            Aircraft.Engine.REFERENCE_SLS_THRUST, units='lbf'))

        scaled_sls_thrust = np.array(inputs[Aircraft.Engine.SCALED_SLS_THRUST])

        thrust_ratio = scaled_sls_thrust / ref_sls_thrust

        # if the engine mass is not scaled, derivatives default to zero
        thrust_deriv = np.zeros(count, dtype=scaled_sls_thrust.dtype)

        # engine mass derivatives
        # indices where scaling is applied
        scale_idx = np.where(scale_mass)
        # indices where scaling is applied and scaling equation is used
        param_idx = np.where(scaling_parameter[scale_idx] >= 0.3)

        thrust_deriv[scale_idx] = scaling_parameter[scale_idx]
        scaled_mass = ref_engine_mass[param_idx] * \
            thrust_ratio[param_idx]**scaling_parameter[param_idx]
        thrust_deriv[param_idx] = (scaling_parameter[param_idx] *
                                   scaled_mass) / scaled_sls_thrust[param_idx]

        J[Aircraft.Engine.MASS,
            Aircraft.Engine.SCALED_SLS_THRUST
          ] = thrust_deriv

        J[Aircraft.Propulsion.TOTAL_ENGINE_MASS,
            Aircraft.Engine.SCALED_SLS_THRUST
          ] = thrust_deriv * num_engines

        J[Aircraft.Engine.ADDITIONAL_MASS,
            Aircraft.Engine.SCALED_SLS_THRUST
          ] = addtl_mass_fraction * thrust_deriv
